"""
All configurations for pymakehelper
"""


from pytconf import Config, ParamCreator


class ConfigVerbose(Config):
    """
    Parameters for verbosity
    """
    verbose = ParamCreator.create_bool(
        help_string="be verbose?",
        default=False,
    )


class ConfigSymlinkInstall(Config):
    """
    Parameters for the symlink install tool
    """
    source_folder = ParamCreator.create_existing_folder(
        help_string="Which folder to install from?",
    )
    target_folder = ParamCreator.create_existing_folder(
        help_string="Which folder to install to?",
    )
    recurse = ParamCreator.create_bool(
        help_string="should I recurse?",
        default=True,
    )
    doit = ParamCreator.create_bool(
        help_string="actually perform the actions?",
        default=True,
    )
    debug = ParamCreator.create_bool(
        help_string="print what we are doing?",
        default=True,
    )
    unlink = ParamCreator.create_bool(
        help_string="remove target files if they are links?",
        default=False,
    )
    incremental = ParamCreator.create_bool(
        help_string="Only do new links?",
        default=True,
    )
    unlink_all = ParamCreator.create_bool(
        help_string="first unlink all targets?",
        default=False,
    )


class ConfigPdflatex(Config):
    """
    Parameters for pdflatex
    """
    input_file = ParamCreator.create_existing_file(
        help_string="input file",
    )
    output_file = ParamCreator.create_new_file(
        help_string="output file",
    )
    remove_tmp = ParamCreator.create_bool(
        help_string="remove the tmp file for output at the end of the run?",
        default=True,
    )
    qpdf = ParamCreator.create_bool(
        help_string="do you want to run the qpdf post processing stage?",
        default=True,
    )
    runs = ParamCreator.create_int(
        help_string="how many times to run pdflatex(1)?",
        default=2,
    )



""" version which can be consumed from within the module """
VERSION_STR = "0.0.38"
DESCRIPTION = "pymakehelper eases writing recipes when using the make system"
APP_NAME = "pymakehelper"
LOGGER_NAME = "pymakehelper"



"""
main file
"""

import os  # for walk, symlink, listdir, unlink, mkdir
import os.path  # for join, expanduser, realpath, abspath, islink, isdir, isfile
import subprocess
import sys

import pylogconf.core
from pytconf import register_endpoint, get_free_args, config_arg_parse_and_launch, \
    register_main

from pymakehelper.configs import ConfigSymlinkInstall, ConfigVerbose
from pymakehelper.static import DESCRIPTION, APP_NAME, VERSION_STR
from pymakehelper.utils import touch_mkdir_many, get_logger, do_install, file_gen, get_flags
from pymakehelper.subprocess import run_error_on_print_or_error, run_no_err, run_only_print_on_error
from pymakehelper.wrapper_pdflatex import run_wrapper_pdflatex
from pymakehelper.configs import ConfigPdflatex
from pymakehelper.subprocess import run_error_on_print


@register_endpoint(
    description="Run pdflatex",
    configs=[
        ConfigPdflatex,
        ConfigVerbose,
    ],
)
def wrapper_pdflatex() -> None:
    run_wrapper_pdflatex()


@register_endpoint(
    description="Install symlinks to things in a folder",
    configs=[
        ConfigSymlinkInstall,
        ConfigVerbose,
    ],
)
def symlink_install() -> None:
    logger = get_logger()
    # first unlink all paths in target leading back to here
    if ConfigSymlinkInstall.unlink_all:
        if os.path.isdir(ConfigSymlinkInstall.target_folder):
            for filename in os.listdir(ConfigSymlinkInstall.target_folder):
                full = os.path.join(ConfigSymlinkInstall.target_folder, filename)
                if os.path.islink(full):
                    link_target = os.path.realpath(full)
                    if link_target.startswith(ConfigSymlinkInstall.source_folder):
                        if ConfigSymlinkInstall.doit:
                            logger.info(f"unlinking [{full}]")
                            os.unlink(full)
    if not os.path.isdir(ConfigSymlinkInstall.target_folder):
        os.mkdir(ConfigSymlinkInstall.target_folder)
    # convert to abs path
    if os.path.isabs(ConfigSymlinkInstall.source_folder):
        source_folder = ConfigSymlinkInstall.source_folder
    else:
        source_folder = os.path.abspath(ConfigSymlinkInstall.source_folder)
    if os.path.isabs(ConfigSymlinkInstall.target_folder):
        target_folder = ConfigSymlinkInstall.target_folder
    else:
        target_folder = os.path.abspath(ConfigSymlinkInstall.target_folder)
    # now create the new links
    for dirname, directories, files in file_gen(source_folder, ConfigSymlinkInstall.recurse):
        for file in files:
            source = os.path.join(dirname, file)
            diff = source[len(source_folder) + 1:]
            target = os.path.join(target_folder, diff)
            do_install(source, target)
        for directory in directories:
            source = os.path.abspath(os.path.join(dirname, directory))
            diff = source[len(source_folder) + 1:]
            target = os.path.join(target_folder, diff)
            if os.path.exists(target):
                assert os.path.isdir(target)
            else:
                logger.info(f"os.mkdir [{target}]")
                os.mkdir(target)
            # do_install(source, target)


@register_endpoint(
    description="Remove one folder from each file name",
    allow_free_args=True,
)
def remove_folders() -> None:
    result = []
    for filename in get_free_args():
        r = os.sep.join(filename.split(os.sep)[1:])
        result.append(r)
    print(" ".join(result), end="")


@register_endpoint(
    description="Touch a file, possibly creating folders for it",
    allow_free_args=True,
)
def touch_mkdir() -> None:
    touch_mkdir_many(get_free_args())


@register_endpoint(
    description="Return with no error even if underlying process returns error",
    allow_free_args=True,
)
def no_err() -> None:
    run_no_err(get_free_args())


@register_endpoint(
    description="Collect stdout and stderr and print them only when the program returns error",
    allow_free_args=True,
    min_free_args=1,
    configs=[
        ConfigVerbose,
    ],
)
def only_print_on_error() -> None:
    run_only_print_on_error(get_free_args())


@register_endpoint(
    description="disregard exit code of command and issue error if there is stdout or stderr output",
    allow_free_args=True,
    min_free_args=1,
    configs=[
        ConfigVerbose,
    ],
)
def error_on_print() -> None:
    run_error_on_print(get_free_args())


@register_endpoint(
    description="if the programin returns an error or prints anything issue an error print",
    allow_free_args=True,
    min_free_args=1,
    configs=[
        ConfigVerbose,
    ],
)
def error_on_print_or_error() -> None:
    run_error_on_print_or_error(get_free_args())


@register_endpoint(
    description="reverse the exitcode of a command",
    allow_free_args=True,
    min_free_args=1,
    configs=[
        ConfigVerbose,
    ],
)
def reverse_exitcode() -> None:
    if ConfigVerbose.verbose:
        print(" ".join(get_free_args()))
    exitcode = subprocess.call(get_free_args())
    sys.exit(int(not bool(exitcode)))


@register_endpoint(
    description="run a build",
    allow_free_args=True,
    min_free_args=0,
)
def run_make() -> None:
    if not os.path.isfile("Makefile"):
        return
    if os.path.isfile(".build.disable"):
        return
    args = ["make"]
    args.extend(get_free_args())
    subprocess.check_call(args)


@register_endpoint(
    description="run command with options of ignore",
    allow_free_args=True,
    min_free_args=3,
)
def run_with_ignore() -> None:
    args = get_free_args()
    filename = args[0]
    flag = args[1]
    cmd = args[2:]
    flags = get_flags(filename)
    if flag in flags:
        sys.exit(0)
    subprocess.check_call(cmd)


@register_endpoint(
    description="wrap css validation command line",
    allow_free_args=True,
    min_free_args=1,
)
def wrapper_css_validator() -> None:
    args = get_free_args()
    out = subprocess.check_output(args).decode()
    do_print = False
    error = False
    for line in out.split("\n"):
        if line.startswith("Sorry"):
            do_print = True
            error = True
        if line.startswith("Valid"):
            do_print = False
        if do_print:
            print(line)
    if error:
        sys.exit(1)


@register_endpoint(
    description="print the environment variables pymakehelper is using",
)
def env() -> None:
    print(os.environ)


@register_main(
    main_description=DESCRIPTION,
    app_name=APP_NAME,
    version=VERSION_STR,
)
def main():
    pylogconf.core.setup()
    config_arg_parse_and_launch()


if __name__ == "__main__":
    main()



"""
utils.py
"""

import logging
import sys
import os

from pymakehelper.static import APP_NAME
from pymakehelper.configs import ConfigSymlinkInstall
from pymakehelper.configs import ConfigVerbose


def unlink_check(filename: str):
    """
    this is a function that removes a file and can optionally die if there is a problem
    """
    if ConfigVerbose.verbose:
        print(f"unlinking [{filename}]", file=sys.stderr)
    if os.path.isfile(filename):
        os.unlink(filename)


def ensure_dir(f, exist_ok=True):
    folder = os.path.dirname(f)
    if folder != "":
        os.makedirs(folder, exist_ok=exist_ok)


def touch(f):
    if os.path.isfile(f):
        os.utime(f, None)
    else:
        with open(f, "w"):
            pass


def touch_mkdir(f):
    ensure_dir(f)
    touch(f)


def touch_mkdir_many(filenames):
    for filename in filenames:
        touch_mkdir(filename)


def get_logger():
    return logging.getLogger(APP_NAME)


def do_real_install(source, target):
    """ This does the real installation """
    logger = get_logger()
    if ConfigSymlinkInstall.unlink:
        if os.path.islink(target):
            logger.info(f"unlinking [{target}]")
            os.unlink(target)
    if ConfigSymlinkInstall.doit:
        logger.info(f"symlinking [{source}], [{target}]")
        os.symlink(source, target)


def do_install(source, target):
    """install a single symlink"""
    logger = get_logger()
    if ConfigSymlinkInstall.incremental:
        if os.path.exists(target):
            assert os.path.islink(target)
            if os.readlink(target) != source:
                logger.info(f"unlinking [{target}]")
                os.unlink(target)
                do_real_install(source, target)
        else:
            do_real_install(source, target)
    else:
        do_real_install(source, target)


def file_gen(root_folder: str, recurse: bool):
    """generate all files in a folder"""
    if recurse:
        yield from os.walk(root_folder)
    else:
        directories = []
        files = []
        for file in os.listdir(root_folder):
            full = os.path.join(root_folder, file)
            if os.path.isdir(full):
                directories.append(file)
            if os.path.isfile(full):
                files.append(file)
        yield root_folder, directories, files


def get_flags(filename: str) -> set[str]:
    """ read the content of a file and get all the flags declared in it """
    flags = set()
    with open(filename, encoding="utf-8") as stream:
        for line in stream:
            line = line.strip()
            if line.find("FLAGS:") != -1:
                all_flags = line.split(":")[1]
                for flag in all_flags.split(","):
                    flags.add(flag.strip())
    return flags



"""
This is a script that runs pdflatex for us.
Why do we need this script ?
- to remove the output before we run pdflatex so that we will be sure that we start clean.
if pdflatex finds a file it will * reprocess * it and we dont want that, do we ?
- we need to run pdflatex twice to create indexes and more.
- pdf latex is way too verbose - we want to remove that output and only show it if there is
an error.
- in case we fail we want to make sure we remove the output.

Maybe more reasons will follow...

Take note of the argument we pass to pdflatex:
- -interaction=nonstopmode - this means that latex will not stop and enter interactive
mode to ask the user what to do about an error (what is this behaviour anyway ?!?).
- -halt-on-error - this means that latex will stop on error.
- -output-directory - this tells pdflatex where the output folder is.

This python script is a rewrite of a similar script in perl.
"""

import sys
import os
import os.path
from pymakehelper.configs import ConfigPdflatex, ConfigVerbose
from pymakehelper.subprocess import run_only_print_on_error
from pymakehelper.utils import unlink_check


def printout(filename: str):
    """
    print to stdout a file content
    this function is adjusted for the ugly output that pdflatex produces and so it
    only prints the lines between lines starting with "!" (including the actual lines
    starting with "!"). Apparently this is how pdflatex shows errors. Ugrrr...
    """
    if ConfigVerbose.verbose:
        print(f"printing [{filename}]", file=sys.stderr)
    with open(filename, encoding="UTF8") as file:
        inerr = False
        for line in file:
            if inerr:
                print(line, file=sys.stderr)
                inerr = False
            else:
                if line.startswith("!"):
                    print(line, file=sys.stderr)
                    inerr = True


def chmod_check(filename: str, check: bool):
    """
    this is a function that chmods a file and can optionally die if there is a problem
    """
    if ConfigVerbose.verbose:
        print(f"chmodding [{filename}]", file=sys.stderr)
    if check:
        os.chmod(filename, 0o444)
    else:
        try:
            os.chmod(filename, 0o444)
        # pylint: disable=broad-exception-caught
        except Exception:
            pass


def my_rename(old_filename: str, new_filename: str, check: bool):
    """
    this is a function that renames a file and dies if there is a problem
    """
    if ConfigVerbose.verbose:
        print(f"my_rename [{old_filename, new_filename}]", file=sys.stderr)
    if check:
        os.rename(old_filename, new_filename)
    else:
        try:
            os.rename(old_filename, new_filename)
        # pylint: disable=broad-exception-caught
        except Exception:
            pass


def run_wrapper_pdflatex():
    """ main entry point """
    filename_input = ConfigPdflatex.input_file
    filename_output = ConfigPdflatex.output_file
    output_dir = os.path.dirname(filename_output)
    output_base = os.path.splitext(filename_output)[0]

    args = [
        "pdflatex",
        "-shell-escape",
        "-interaction=nonstopmode",
        "-halt-on-error",
        "-output-directory",
        output_dir,
        filename_input,
    ]
    if ConfigVerbose.verbose:
        print(f"input is [{filename_input}]")
        print(f"output is [{filename_output}")
        print(f"cmd is [{args}")
    # first remove the output (if it exists)
    unlink_check(
        filename_output,
    )
    # we need to run the command twice!!! (to generate the index and more)
    for _ in range(ConfigPdflatex.runs):
        run_only_print_on_error(args)
        unlink_check(output_base + ".log")
        unlink_check(output_base + ".out")
        unlink_check(output_base + ".toc")
        unlink_check(output_base + ".aux")
        unlink_check(output_base + ".nav")
        unlink_check(output_base + ".snm")
        unlink_check(output_base + ".vrb")

    if ConfigPdflatex.qpdf:
        # move the output to the new place
        tmp_output = filename_output + ".pdf"
        my_rename(filename_output, tmp_output, True)
        # I also had "--force-version=1.5" but it is not needed since I use
        # pdflatex and pdftex with the right version there...
        args = [
            "qpdf",
            "--deterministic-id",
            "--linearize",
            tmp_output,
            filename_output,
        ]
        run_only_print_on_error(args)
        unlink_check(tmp_output)



"""
This module provides a simple wrapper around the subprocess module
"""

import sys
import subprocess
from pymakehelper.configs import ConfigVerbose


def run_no_err(args: list[str]):
    if ConfigVerbose.verbose:
        print(" ".join(args))
    subprocess.call(args)


def run_only_print_on_error(args: list[str]):
    if ConfigVerbose.verbose:
        print(" ".join(args))
    with subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as pr:
        out_out, out_err = pr.communicate()
        code = pr.returncode
    if code:
        print(out_out.decode(), end="", file=sys.stdout)
        print(out_err.decode(), end="", file=sys.stderr)
        sys.exit(code)


def run_error_on_print(args: list[str]):
    if ConfigVerbose.verbose:
        print(" ".join(args))
    with subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as pr:
        out_out, out_err = pr.communicate()
        _ = pr.returncode
    if len(out_out) > 0 or len(out_err) > 0:
        print(out_out.decode(), end="", file=sys.stdout)
        print(out_err.decode(), end="", file=sys.stderr)
        sys.exit(1)
    else:
        sys.exit(0)


def run_error_on_print_or_error(args: list[str]):
    if ConfigVerbose.verbose:
        print(" ".join(args))
    with subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as pr:
        out_out, out_err = pr.communicate()
        code = pr.returncode
    if len(out_out) > 0 or len(out_err) > 0 or code:
        print(out_out.decode(), end="", file=sys.stdout)
        print(out_err.decode(), end="", file=sys.stderr)
        sys.exit(1)
    else:
        sys.exit(0)
